/*
 * Copyright (C) 2014-2019 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <linux/list.h>
#include <linux/hrtimer.h>

#include "../include/fa-board.h"
#include "fa-drv.h"
#include "fa-pci.h"
#include "fa-fip.h"

#define BOARD_INFO "FIP/WorldFIP Arbitrator board"

static struct class *cl = NULL; // Global variable for the device class
int pci_major = PCI_MAJOR;
int pci_devs  = PCI_DEVS;
static u8 minors[PCI_DEVS] = {0};
static LIST_HEAD(privates);

MODULE_DESCRIPTION("FIP/WorldFIP Arbitrator driver for PCI/PCIe board [FIPCode6]");
MODULE_AUTHOR("Exoligent SARL.");
MODULE_LICENSE("GPL");

MODULE_VERSION("2.5");

static struct pci_device_id ids[] = {   {PCI_DEVICE(PCI_VENDOR_ID_PLX, PCI_DEV_ID_PLX_9030),},
                                        {0,}};

static struct {const char *name;} board_info[] = { {BOARD_INFO},};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
    static irqreturn_t irq_handler(int irq, void *dev_id)
#else
    static irqreturn_t irq_handler(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
    u32 intcsr_reg;
    u32 limit_u32 = -1;
    ktime_t currtime;
    s64 cycle_time;
    struct pci_priv_device *pdev = (struct pci_priv_device *) dev_id;

    intcsr_reg = read_register_u32(pdev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR);

    /* is the IRQ signal is produced by the device ? */
    if ((intcsr_reg & C_MK_REG_BIT_IT_LINT1_ACTIVE) ||
        (intcsr_reg & C_MK_REG_BIT_IT_LINT2_ACTIVE)) {
        /* irq produced by this device */
        /* disable IRQ */
        intcsr_reg &= ~C_MK_REG_BIT_IT_ENABLE;
        write_register_u32(pdev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR, intcsr_reg);

        pdev->linti1 = 0;
        pdev->linti2 = 0;

        if (intcsr_reg & C_MK_REG_BIT_IT_LINT1_ACTIVE)
            pdev->linti1 = 1;

        if (intcsr_reg & C_MK_REG_BIT_IT_LINT2_ACTIVE) {
            pdev->linti2 = 1;
            pdev->stat.cycle_count++;

            /* (eoc) cycle period calculation */
            currtime = ktime_get();
            cycle_time = ktime_to_ns(ktime_sub(currtime, pdev->prectime));
            pdev->prectime = currtime;
            if (cycle_time > 0)
                pdev->stat.cycle_time = cycle_time;

            reset_eoc_interrupt(pdev);
        }

        /* raise the interrupt flag */
        pdev->int_flag = 1;

        /* wake up the mutex */
        wake_up(&pdev->int_mutex);

        (pdev->irq_handled_counter == limit_u32) ? 0 : pdev->irq_handled_counter++;
        return IRQ_HANDLED;
    } else {
        /* irq not produced by this device */
        (pdev->irq_none_handled_counter == limit_u32) ? 0 : pdev->irq_none_handled_counter++;
        return IRQ_HANDLED;
    }
}

static void pci_setup_cdev(struct pci_priv_device* dev, int index)
{
    char dev_name[64];
    int err, devno = MKDEV(pci_major, index);

    /* add class to device - ls /dev/ */
    sprintf(dev_name, "%s%d", DRV_NAME, index);
    dev->dev = device_create(cl, NULL, devno, NULL, dev_name);
    if (IS_ERR(dev->dev))
        printk(KERN_ERR DRV_HDR "Error when adding class to device (index %d)", index);

    cdev_init(&dev->cdev, &pcifip_fops);
    dev->cdev.owner = THIS_MODULE;
    dev->cdev.ops = &pcifip_fops;
    err = cdev_add (&dev->cdev, devno, pci_devs);
    if (err)
        printk(KERN_ERR DRV_HDR "Error %d when adding device (index %d)", err, index);
}

static int probe(struct pci_dev *dev, const struct pci_device_id *id)
{
    int rc = 0;
    struct pci_priv_device *priv_dev;
    int i;
    u32 local0;
    u32 local1;
    void *mmio[DEVICE_COUNT_RESOURCE];   /* Remaped I/O memory */
    u32 mmio_len[DEVICE_COUNT_RESOURCE]; /* Size of remaped I/O memory */
    u16 revision;
    u8 irq_pin;
    u8 ioRessourceMem;
    u8 free_minor_dectected = 0;
    unsigned long flag;

    spinlock_t to_lock = __SPIN_LOCK_UNLOCKED(to_lock);
    int err            = pci_enable_device(dev);
    if (err)
        return err;

    printk(KERN_DEBUG DRV_HDR "==> Device probing: %s\n", 
        board_info[id->driver_data].name);
    
    /* Allocate a private structure and reference it as driver's data */
    priv_dev = (struct pci_priv_device *) kcalloc(1, 
                        sizeof(struct pci_priv_device), GFP_KERNEL);
    if (priv_dev == NULL)
        return -ENODEV;

    priv_dev->pci_dev = dev;

    spin_lock_init(&to_lock);
    spin_lock(&to_lock);

    for (i = 0; i < PCI_DEVS; i++) {
        if (minors[i] == 0) {
            minors[i] = 1 ;
            priv_dev->type = i;
            free_minor_dectected = 1;
            break;
        }
    }

    if (free_minor_dectected == 0)
        return -EFAULT;

    spin_unlock(&to_lock);
    
    priv_dev->minor = priv_dev->type ;

    pci_set_drvdata(dev, priv_dev);
    pci_setup_cdev(priv_dev, priv_dev->minor);
    pci_read_config_dword(dev,0x18, &local0);
    pci_read_config_dword(dev,0x1C, &local1);

    /* Reserve PCI I/O and memory resources */
    rc = pci_request_regions(dev, DRV_NAME);
    if (rc)
        return -ENODEV;

    ioRessourceMem = 0;
    for (i = 0; i < DEVICE_COUNT_RESOURCE; i++)  {

        if (pci_resource_start(dev, i) != 0) {
            printk(KERN_DEBUG DRV_HDR "    BAR %d (%#08x - %#08x), len = %d, flags = %#08x\n", i,
                (u32) pci_resource_start(dev, i),
                (u32) pci_resource_end(dev, i),
                (u32) pci_resource_len(dev, i),
                (u32) pci_resource_flags(dev, i));
        }
        
        flag = pci_resource_flags(dev, i);

        if ((flag & IORESOURCE_MEM) == IORESOURCE_MEM) {

            mmio[i] = ioremap(pci_resource_start(dev, i), pci_resource_len(dev, i));
            if (mmio[i] == NULL) {
                rc = -ENOMEM;
                return rc;
            }
    
            mmio_len[i] = pci_resource_len(dev, i);

            if (ioRessourceMem == 0) {
                priv_dev->IoSpacePciPlx.bSpaceMapped         = true;
                priv_dev->IoSpacePciPlx.nSpaceRange          = mmio_len[i];
                priv_dev->IoSpacePciPlx.pSpaceBase           = mmio[i];
                priv_dev->IoSpacePciPlx.physicalAddress      = pci_resource_start(dev, i);
            } else if (ioRessourceMem == 1) { 
                priv_dev->IoSpacePciFullFip2.bSpaceMapped    = true;
                priv_dev->IoSpacePciFullFip2.nSpaceRange     = mmio_len[i];
                priv_dev->IoSpacePciFullFip2.pSpaceBase      = mmio[i];
                priv_dev->IoSpacePciFullFip2.physicalAddress = pci_resource_start(dev, i);
            } else if (ioRessourceMem == 2) {  
                priv_dev->IoSpacePciFieldual.bSpaceMapped    = true;
                priv_dev->IoSpacePciFieldual.nSpaceRange     = mmio_len[i];
                priv_dev->IoSpacePciFieldual.pSpaceBase      = mmio[i];
                priv_dev->IoSpacePciFieldual.physicalAddress = pci_resource_start(dev, i);
            } else if (ioRessourceMem == 3) {
                priv_dev->IoSpacePciRam.bSpaceMapped         = true;
                priv_dev->IoSpacePciRam.nSpaceRange          = mmio_len[i];
                priv_dev->IoSpacePciRam.pSpaceBase           = mmio[i];
                priv_dev->IoSpacePciRam.physicalAddress      = pci_resource_start(dev, i);
            }
            ioRessourceMem++;
        } else {
            mmio[i] = NULL;
            if ((flag & IORESOURCE_IO) == IORESOURCE_IO) {
                priv_dev->IoPortPciPlxAddress = pci_resource_start(dev, i);
                priv_dev->IoPortPciPlxSize = pci_resource_len(dev, i);
            }
        }
    }

    /* map bus local registers addresses (fullfip/fieldual) */
    init_fip_infos(priv_dev);
    printk(KERN_DEBUG DRV_HDR "    Fullfip/Fieldual registers offset = %d\n", 
                priv_dev->reg_offset_factor);


    rc = pci_read_config_word(dev, PCI_VENDOR_ID, &revision);
    if (rc)
        return rc;

    /* Lock init */
    spin_lock_init(&priv_dev->lock);

    rc = pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &irq_pin);
    if (rc)
        return rc;

    printk(KERN_DEBUG DRV_HDR "    IRQ pin = #%d (0=none, 1=INTA#...4=INTD#).\n", irq_pin);
    printk(KERN_DEBUG DRV_HDR "    IRQ line = #%d.\n", dev->irq);

    /* Init INTCSR register */
    init_interrupt(priv_dev);

    /* Register IRQ */
    priv_dev->int_flag = 0;
    init_waitqueue_head (&priv_dev->int_mutex);

    rc = request_irq(dev->irq, &irq_handler, IRQF_SHARED, DRV_NAME, priv_dev);
    if (rc) {
        priv_dev->irq_line = -1;
        return rc;
    }
    priv_dev->irq_line = (int) dev->irq;

    /* Reset IRQ counter's */
    priv_dev->irq_handled_counter = 0;
    priv_dev->irq_none_handled_counter = 0;

    spin_lock(&to_lock);
    list_add_tail(&priv_dev->link, &privates);
    spin_unlock(&to_lock);

    return 0;
}

static void remove(struct pci_dev *dev)
{
    spinlock_t to_lock = __SPIN_LOCK_UNLOCKED(to_lock);
    u32 intcsr;
    
    struct pci_priv_device *priv_dev = pci_get_drvdata(dev);
    printk(KERN_DEBUG DRV_HDR "==> Device removing (idx=%d)\n", priv_dev->minor);

    /* free IRQ */
    if (priv_dev->irq_line >= 0) {
        intcsr = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR);
        intcsr &= ~C_MK_REG_BIT_IT_ENABLE;
        write_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR, intcsr);

        free_irq(priv_dev->irq_line, (void *) priv_dev);
    }

    if (priv_dev->IoSpacePciPlx.bSpaceMapped)
        iounmap(priv_dev->IoSpacePciPlx.pSpaceBase);

    if (priv_dev->IoSpacePciFullFip2.bSpaceMapped)
        iounmap(priv_dev->IoSpacePciFullFip2.pSpaceBase);

    if (priv_dev->IoSpacePciFieldual.bSpaceMapped)
        iounmap(priv_dev->IoSpacePciFieldual.pSpaceBase);

    spin_lock_init(&to_lock);

    spin_lock(&to_lock);
    minors[priv_dev->minor] = 0;
    spin_unlock(&to_lock);

    pci_disable_device(dev);
    pci_release_regions(dev);

    /* device destroying */
    if (cl) {
        int devno = MKDEV(pci_major, priv_dev->minor);
        device_destroy(cl, devno);
    }

    cdev_del(&priv_dev->cdev);

    spin_lock(&to_lock);
    list_del(&priv_dev->link);
    spin_unlock(&to_lock);

    kfree(priv_dev);
}

static struct pci_driver pci_driver = {
                                        .name       = DRV_NAME,
                                        .id_table   = ids,
                                        .probe      = probe,
                                        .remove     = remove
                                    };

MODULE_DEVICE_TABLE (pci, ids);

static int __init pci_init_module(void)
{
    int err;
    dev_t dev;

    printk(KERN_DEBUG DRV_HDR "==> Init: %s\n", board_info[0].name);

    /* create device class - ls /sys/class */
    cl = class_create(THIS_MODULE, DRV_NAME);

    /* allocate device numbers - cat /proc/devices */
    err = alloc_chrdev_region(&dev, 0, pci_devs, DRV_NAME);
    if (!err) {
        pci_major = MAJOR(dev);

        /* registering driver */
        err = pci_register_driver(&pci_driver);
    }
    return err;
}

static void __exit pci_exit_module(void)
{
    printk(KERN_DEBUG DRV_HDR "==> Exit: %s\n", board_info[0].name);

    pci_unregister_driver(&pci_driver);

    if (cl)
        class_destroy(cl);
    
    unregister_chrdev_region(MKDEV(pci_major, 0), pci_devs);
}

ssize_t pcifip_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
    int real;
    u8* ptr;
    int i;
    struct pci_priv_device* priv_dev = file->private_data;

    if (priv_dev == 0)
        return -1;

    /* Check for overflow */
    if (count <= priv_dev->IoSpacePciPlx.nSpaceRange - (int) *ppos)
        real = count;
    else
        real = priv_dev->IoSpacePciPlx.nSpaceRange - (int) *ppos;
    
    ptr = (u8*) priv_dev->IoSpacePciPlx.pSpaceBase;
    
    if (real) {
        for (i = 0; i < real; i++) {
            int error;
            u8 c =(u8) ioread8((u8*) ptr);
            printk(KERN_DEBUG "%x ", c);
            ptr++;
            error = put_user(c, buf);
            if (error)
                break;
            (char*) buf++;
        }   
        *ppos += real;
    }
    return real;
}

ssize_t pcifip_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    int real;
    struct pci_priv_device *priv_dev = file->private_data;

    if (priv_dev == 0)
        return 0;

    /* Check for overflow */
    if (count <= priv_dev->IoSpacePciPlx.nSpaceRange - (int)*ppos)
        real = count;
    else
        real = priv_dev->IoSpacePciPlx.nSpaceRange - (int)*ppos;

    if (real)
        copy_from_user((char*)priv_dev->IoSpacePciPlx.pSpaceBase + (int)*ppos, buf, real);

    *ppos += real;

    return real;
}

int pcifip_open(struct inode *inode, struct file *file)
{
    int minor = MINOR(inode->i_rdev);
    struct list_head *cur;
    struct pci_priv_device *priv_dev;

    list_for_each(cur, &privates)
    {
        priv_dev = list_entry(cur, struct pci_priv_device, link);
        if (priv_dev->minor == minor) {
            if (priv_dev->is_open)
                return -EBUSY;

            init_fip_infos(priv_dev);
            init_interrupt(priv_dev);
            file->private_data = priv_dev;
            priv_dev->is_open = 1;
            return 0;
        }
    }
    return -ENODEV;
}


int pcifip_release(struct inode *inode, struct file *file)
{
    struct pci_priv_device *priv_dev = file->private_data;
    if (file->private_data == NULL)
        return 0;

    priv_dev->is_open = 0;
    file->private_data = 0;
    return 0;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35))
    int pcifip_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
#else
    long pcifip_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#endif
{
    struct pci_priv_device *priv_dev;
    void* ibuf_ptr = NULL;
    void* obuf_ptr = NULL;
    u32 ibuf_sz = 0, obuf_sz = 0;
    u32 status, err, io_status = 0;
    struct arg_io args;
    struct arg_io __user *args_ptr = 0;
    
    status = STATUS_SUCCESS;
    io_status = 0;

    if (arg != 0) {

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(arg, sizeof(u32*)))
#else
        if (!access_ok(VERIFY_WRITE, arg, sizeof(u32*)))
#endif
            return -EFAULT;

        if (copy_from_user(&args, (void*) arg, sizeof(args)))
            return -EFAULT;

        args_ptr = (struct arg_io __user *) arg;
        ibuf_sz  = args.lgIn;
        obuf_sz = args.lgOut;

        if (ibuf_sz) {
            ibuf_ptr = kmalloc(ibuf_sz * sizeof(u8), GFP_KERNEL);
            err = copy_from_user(ibuf_ptr, args.argIO_In, ibuf_sz);
            if (err)
                return -EFAULT;
        } 

        obuf_ptr = args.argIO_Out;
    }

    priv_dev = file->private_data;
    
    switch (cmd) 
    {
        case IOCTL_EDA_GET_MONITORING_INFOS: // O
            {
                EDA_GET_MONITORING_INFOS_OUTPUT obuf = {0};
                u32 registerValue = 0;
                u8 magicNumber = 0;

                if (obuf_sz != sizeof(EDA_GET_MONITORING_INFOS_OUTPUT)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                // by default: -256°C (2's complement) ==> non significant value
                obuf.board_temperature = 0xf000;

                registerValue = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase,
                                    C_REG_PLX_RESOURCES);

                magicNumber = (u8) ((registerValue >> 24) & 0x000000ff);

                if (magicNumber == C_REG_MAGIC_NB) {
                    registerValue = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase,
                                        C_REG_PLX_TEMPERATURE);
                    obuf.board_temperature = (s16) (registerValue & C_MK_REG_TEMP_VALUE);
                }

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(EDA_GET_MONITORING_INFOS_OUTPUT)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(EDA_GET_MONITORING_INFOS_OUTPUT)))
#endif
                    return -EFAULT;

                if (copy_to_user((u8*) args.argIO_Out, (u8*) &obuf, sizeof(EDA_GET_MONITORING_INFOS_OUTPUT)))
                    return -EFAULT;

                io_status = sizeof(EDA_GET_MONITORING_INFOS_OUTPUT);
                status = STATUS_SUCCESS;
            }
            break;
        // IOCTL EDA (ExoDevice Access)
        // --- Management
        case IOCTL_EDA_GET_DEV_INFO: // O
            {
                EDA_DEV_INFO_OUTPUT obuf = {0};

                if (obuf_sz != sizeof(EDA_DEV_INFO_OUTPUT)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                // ... fill here the structure
                obuf.nb_fullfip     = priv_dev->nb_fullfip;
                obuf.nb_fullfip_ram = priv_dev->nb_fullfip_ram;
                obuf.nb_fieldual    = priv_dev->nb_fieldual;

                obuf.plx_base = (unsigned long) priv_dev->IoSpacePciPlx.physicalAddress;
                obuf.plx_size = priv_dev->IoSpacePciPlx.nSpaceRange;
                obuf.fullfip_base =
                        (unsigned long) priv_dev->IoSpacePciFullFip2.physicalAddress;
                obuf.fullfip_size = priv_dev->IoSpacePciFullFip2.nSpaceRange;
                obuf.fullfip_ram_base =
                        (unsigned long) priv_dev->IoSpacePciRam.physicalAddress;
                obuf.fullfip_ram_size = priv_dev->IoSpacePciRam.nSpaceRange;
                obuf.fieldual_base =
                        (unsigned long) priv_dev->IoSpacePciFieldual.physicalAddress;
                obuf.fieldual_size = priv_dev->IoSpacePciFieldual.nSpaceRange;

                obuf.vid = priv_dev->pci_dev->vendor;
                obuf.did = priv_dev->pci_dev->device;
                obuf.ssvid = priv_dev->pci_dev->subsystem_vendor;
                obuf.ssid = priv_dev->pci_dev->subsystem_device;
                obuf.irq_number = (int) priv_dev->pci_dev->irq;
                obuf.regs_offset_factor = (int) priv_dev->reg_offset_factor;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(EDA_DEV_INFO_OUTPUT)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(EDA_DEV_INFO_OUTPUT)))
#endif
                    return -EFAULT;

                if (copy_to_user((u8*) args.argIO_Out, (u8*) &obuf, sizeof(EDA_DEV_INFO_OUTPUT)))
                    return -EFAULT;

                io_status = sizeof(EDA_DEV_INFO_OUTPUT);
                status = STATUS_SUCCESS;
            } break;

        case IOCTL_EDA_LIST_INTERNAL_REGISTERS: // O
            {
                u32 plxInternalRegisters[C_PLX_REGISTER_NUMBER];
                u32 plxRegistersNumber;

                if (obuf_sz < C_PLX_REGISTER_NUMBER * sizeof(u32)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                // Display, in the driver debugging environnement, the contents of the PLX internal registers
                // For debug purpose
                for (plxRegistersNumber = 0; plxRegistersNumber < C_PLX_REGISTER_NUMBER; plxRegistersNumber++)
                    plxInternalRegisters[plxRegistersNumber] = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, plxRegistersNumber);

                err = copy_to_user((u32*) args.argIO_Out, plxInternalRegisters, C_PLX_REGISTER_NUMBER * sizeof(u32));

                io_status = C_PLX_REGISTER_NUMBER * sizeof(u32);
                status = STATUS_SUCCESS;
            } break;

        case IOCTL_EDA_RELOAD_INTERNAL_REGISTERS: // none
            {
                u32 registerValue;

                // Reading the PLX control register
                registerValue = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);

                // Enable bit for the reload registers
                registerValue = registerValue | C_MK_REG_BIT_EEPROM_RELOAD_HIGH;

                write_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, registerValue);

                status = STATUS_SUCCESS;

            } break;

        case IOCTL_EDA_PLX_WRITE_REGISTER: // I
            {
                EDA_WRITE_REGISTER_INPUT* pIn;

                if (ibuf_sz != sizeof(EDA_WRITE_REGISTER_INPUT)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                pIn = (EDA_WRITE_REGISTER_INPUT*) ibuf_ptr;

                if ((pIn->RelativeAddress > priv_dev->IoSpacePciPlx.nSpaceRange)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                write_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, pIn->RelativeAddress, pIn->Data);

                status = STATUS_SUCCESS;

            } break;

        case IOCTL_EDA_PLX_READ_REGISTER: // I + O
            {
                u32 addressRelative, data;

                if ((ibuf_sz != sizeof(u32)) ||
                    (obuf_sz != sizeof(u32))) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                addressRelative = *((u32*) ibuf_ptr);

                if (addressRelative > priv_dev->IoSpacePciPlx.nSpaceRange) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                data = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, addressRelative);

                err = copy_to_user((u32*) args.argIO_Out, (u32*) &data, sizeof(u32));

                io_status = sizeof(u32);
                status = STATUS_SUCCESS;

            } break;

        // --- EEPROM Management
        case IOCTL_EDA_CHECK_EEPROM: // O
            {
                u32 EEPROMState = 0;
                u32 registerValue;

                // Check if a blank or programmed serial EEPROM is present
                registerValue = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
                registerValue = registerValue & C_MK_REG_BIT_EEPROM_PRESENT_HIGH;

                // EEPROM presence ?
                if (registerValue == C_MK_REG_BIT_EEPROM_PRESENT_HIGH) {
                    EEPROMState |= C_MASK_EEPROM_PRESENT;

                    // Blank EEPROM ?
                    if (!((priv_dev->pci_dev->subsystem_vendor == C_SUB_VENDOR_ID_DEFAULT_VALUE) &&
                          (priv_dev->pci_dev->subsystem_device == C_SUB_SYSTEM_ID_DEFAULT_VALUE))) {
                        EEPROMState |= C_MASK_EEPROM_NOT_BLANK;
                    }
                }

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(u32)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(u32)))
#endif
                    return -EFAULT;

                if (put_user(EEPROMState, (u32*) args.argIO_Out))
                    return -EFAULT;

                io_status = sizeof(u32);
                status = STATUS_SUCCESS;
            } break;

        case IOCTL_EDA_READ_EEPROM: // I + O
            {
                EDA_READ_EEPROM_INPUT* pIn;
                EDA_READ_EEPROM_OUTPUT obuf = {0};
                u8  countOfBytes = 0;

                if ((ibuf_sz != sizeof(EDA_READ_EEPROM_INPUT)) ||
                    (obuf_sz != sizeof(EDA_READ_EEPROM_OUTPUT))) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                pIn = (EDA_READ_EEPROM_INPUT*) ibuf_ptr;

                if (pIn->CountOfWords > EEPROM_USABLE_SIZE) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }
                    
                countOfBytes = plx_eeprom_read(priv_dev , (u8*) &(obuf.Data[0]), pIn->StartAddress, pIn->CountOfWords);
				obuf.CountOfBytes = countOfBytes;

                if (obuf_sz < countOfBytes) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(EDA_READ_EEPROM_OUTPUT)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(EDA_READ_EEPROM_OUTPUT)))
#endif
                    return -EFAULT;

                if (copy_to_user((u8*) args.argIO_Out, (u8*) &obuf, sizeof(EDA_READ_EEPROM_OUTPUT)))
                    return -EFAULT;

                io_status = sizeof(EDA_READ_EEPROM_OUTPUT);
                status = STATUS_SUCCESS;

            } break;

        // --- Useful I/O access
        case IOCTL_EDA_GET_RGB_LED: // O
            {

                if (obuf_sz != sizeof(u8)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(u8)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(u8)))
#endif
                    return -EFAULT;

                if (put_user(priv_dev->LedsState, (u8*) args.argIO_Out))
                    return -EFAULT;

                io_status = sizeof(u8);
                status = STATUS_SUCCESS;

            } break;

        case IOCTL_EDA_SET_RGB_LED: // I
            {
                u32 reg = 0;
                u8 led = 0;

                if (ibuf_sz != sizeof(u8)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                led = *((u8*) ibuf_ptr);
                reg = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_GPIOC);

                /* led gpio4 active */
                reg = (led & 0x01) ? reg | C_MK_REG_BIT_GPIO4_LED_ON :
                                     reg & ~C_MK_REG_BIT_GPIO4_LED_ON;

                /* led blue active */
                reg = (led & 0x02) ? reg | C_MK_REG_BIT_BLUE_LED_ON :
                                     reg & ~C_MK_REG_BIT_BLUE_LED_ON;

                /* led green active */
                reg = (led & 0x04) ? reg | C_MK_REG_BIT_GREEN_LED_ON :
                                     reg & ~C_MK_REG_BIT_GREEN_LED_ON;

                /* led red active */
                reg = (led & 0x08) ? reg | C_MK_REG_BIT_RED_LED_ON :
                                     reg & ~C_MK_REG_BIT_RED_LED_ON;

                write_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_GPIOC, reg);
                priv_dev->LedsState = led;
                status = STATUS_SUCCESS;
            } break;

        // --- Useful information access
        case IOCTL_EDA_GET_PRODUCT_NUMBER: // O
            {
                u32 productNumber = 0;
                u8 startAddress, length, countOfBytes = 0;
                
                if (!priv_dev->CardConf)
                    init_card_conf(priv_dev);

                if ((priv_dev->CardConf & C_MK_REG_PLX_TYPE) == C_REG_PLX9052_TYPE) { // PLX9052
                    startAddress = 0x3E; // (adressage 16-bit)
                }

                if ((priv_dev->CardConf & C_MK_REG_PLX_TYPE) == C_REG_PLX9030_TYPE) { // PLX9030
                    startAddress = 0x5E; // 0xBC (adressage 8-bit) = 2 x 0x5E (adressage 16-bit)
                }

                length  = 0x02;
                countOfBytes = plx_eeprom_read(priv_dev, (u8*) (&productNumber), startAddress, length);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(u32)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(u32)))
#endif
                    return -EFAULT;

                if (put_user(productNumber, (u32*) args.argIO_Out))
                    return -EFAULT;

                io_status = sizeof(u32);
                status = STATUS_SUCCESS;
                
            } break;

        case IOCTL_EDA_GET_DEVICE_ID: // O
            {
                u16 deviceID = 0;
                u8 startAddress, countOfBytes = 0;
                
                if (obuf_sz != sizeof(u16)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                startAddress = 0x00; // Adresse EEPROM du DeviceID
                countOfBytes = plx_eeprom_read(priv_dev, (u8*) (&deviceID), startAddress, sizeof(u16) / 2);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(u16)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(u16)))
#endif
                    return -EFAULT;

                if (put_user(deviceID, (u16*) args.argIO_Out)) {
                    status = -EFAULT;
                    break;
                }

                io_status = sizeof(u16);
                status = STATUS_SUCCESS;
                
            } break;

        case IOCTL_EDA_GET_CARD_CONF: // O
            {
                u32 cardConf = 0;
                u8 startAddress, countOfBytes = 0;

                if (!priv_dev->CardConf)
                    init_card_conf(priv_dev);

                if (obuf_sz != sizeof(u32)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                startAddress = 0x38; // (adressage 16-bit)
                countOfBytes = plx_eeprom_read(priv_dev, (u8*) (&cardConf), startAddress, sizeof(u32) / 2);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(u32)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(u32)))
#endif
                    return -EFAULT;

                if (put_user(cardConf, (u32*) args.argIO_Out)) {
                    status = -EFAULT;
                    break;
                }

                io_status = sizeof(u32);
                status = STATUS_SUCCESS;

            } break;

        case IOCTL_EDA_GET_SUBSYSTEM_VENDOR_ID: // O
            {
                u16 subSystemVendorID = 0;
                u8 startAddress, countOfBytes = 0;

                if (obuf_sz != sizeof(u16)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                startAddress = 0x07; // (adressage 16-bit)
                countOfBytes = plx_eeprom_read(priv_dev, (u8*) (&subSystemVendorID), startAddress, sizeof(u16) / 2);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(u16)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(u16)))
#endif
                    return -EFAULT;

                if (put_user(subSystemVendorID, (u16*) args.argIO_Out)) {
                    status = -EFAULT;
                    break;
                }

                io_status = sizeof(u16);
                status = STATUS_SUCCESS;

            } break;

        case IOCTL_EDA_GET_SUBSYSTEM_ID: // O
            {
                u16 subSystemID = 0;
                u8 startAddress, countOfBytes = 0;

                if (obuf_sz != sizeof(u16)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                startAddress = 0x06; // (adressage 16-bit)
                countOfBytes = plx_eeprom_read(priv_dev, (u8*) (&subSystemID), startAddress, sizeof(u16) / 2);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(u16)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(u16)))
#endif
                    return -EFAULT;

                if (put_user(subSystemID, (u16*) args.argIO_Out)) {
                    status = -EFAULT;
                    break;
                }

                io_status = sizeof(u16);
                status = STATUS_SUCCESS;
            } break;

        case IOCTL_EDA_GET_REVISION_ID: // O
            {
                u16 revisionID = 0;
                u8 startAddress, countOfBytes = 0;

                if (obuf_sz != sizeof(u16)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                startAddress = 0x05; // (adressage 16-bit)
                countOfBytes = plx_eeprom_read(priv_dev, (u8*) (&revisionID), startAddress, sizeof(u16) / 2);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(u16)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(u16)))
#endif
                    return -EFAULT;

                if (put_user(revisionID, (u16*) args.argIO_Out)) {
                    status = -EFAULT;
                    break;
                }

                io_status = sizeof(u16);
                status = STATUS_SUCCESS;
            } break;

        // --- FullFip access
        case IOCTL_EDA_FULLFIP_RESET: // none
            {
                u32 registerValue;
                unsigned long delay;
                bool bWait;

                // Reading the PLX control register
                registerValue = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);

                // Enable bit for reset FullFip
                registerValue |= C_MK_REG_BIT_LOCAL_RESET_ON;

                write_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, registerValue);

                // Waiting for at least 1,024 ms
                delay = jiffies + msecs_to_jiffies(1);
                bWait = true;

                do {
                    if (time_after(jiffies, delay))
                        bWait = false;
                } while (bWait);

                // Disable bit for reset FullFip
                registerValue &= ~C_MK_REG_BIT_LOCAL_RESET_ON;

                write_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, registerValue);

                // Waiting for at least 100 ms
                delay = jiffies + msecs_to_jiffies(100);
                bWait = true;

                do {
                    if (time_after(jiffies, delay))
                        bWait = false;
                } while (bWait);

                status = STATUS_SUCCESS;

            } break;

        // --- FielDual / MAX3000A access
        case IOCTL_EDA_FIELDUAL_VALID_MEDIUM: // I + O
            status = eda_fieldual_valid_medium(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            break;
        case IOCTL_EDA_FIELDUAL_GET_MEDIUM: // O
            status = eda_fieldual_get_medium(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            break;
        // --- IRQ management
        case IOCTL_EDA_IRQ_WAIT_FOR: // O
            {
                u8 data[2] = {0};

                wait_event(priv_dev->int_mutex, priv_dev->int_flag);

                if (!priv_dev->int_flag) {
                    status = -EINTR;
                    break;
                }
                priv_dev->int_flag = 0;
                data[0] = priv_dev->linti1;
                data[1] = priv_dev->linti2;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(u16)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(u16)))
#endif
                    return -EFAULT;

                if (copy_to_user((u8*) args.argIO_Out, (u8*) data, sizeof(u16)))
                    return -EFAULT;

                io_status = sizeof(u16);
                status = STATUS_SUCCESS;
            } break;

        case IOCTL_EDA_IRQ_REARM: // none
            {
                /* enable IRQ */
                u32 intcsr_reg = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR);
                intcsr_reg |= C_MK_REG_BIT_IT_ENABLE;
                write_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR, intcsr_reg);

                status = STATUS_SUCCESS;
            } break;

        case IOCTL_EDA_IRQ_RESUME: // none
            {
                /* wake up 'manually' the mutex */
                /* useful to unlock a wait on an event that won't come */

                /* disable IRQ */
                u32 intcsr_reg = read_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR);
                intcsr_reg &= ~C_MK_REG_BIT_IT_ENABLE;
                write_register_u32(priv_dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR, intcsr_reg);
                /* raise the interrupt flag */
                priv_dev->int_flag = 1;
                /* wake up the mutex */
                wake_up(&priv_dev->int_mutex);

                status = STATUS_SUCCESS;
            } break;

        case IOCTL_EDA_RESET_STATS: // none
            {
                struct fip_stat stat = {0};
                priv_dev->irq_handled_counter = 0;
                priv_dev->irq_none_handled_counter = 0;
                priv_dev->stat = stat;
                status = STATUS_SUCCESS;
            } break;

        case IOCTL_EDA_GET_STATS: // none
            {
                EDA_GET_STATS_OUTPUT obuf = {0};

                if (obuf_sz != sizeof(EDA_GET_STATS_OUTPUT)) {
                    status = STATUS_INVALID_PARAMETER;
                    break;
                }

                obuf.IrqHandledCount = priv_dev->irq_handled_counter;
                obuf.IrqNonHandledCount = priv_dev->irq_none_handled_counter;
                obuf.CycleCount = priv_dev->stat.cycle_count;
                obuf.CycleTime = priv_dev->stat.cycle_time;
                obuf.VarPSentCount = priv_dev->stat.var_p_sent_count;
                obuf.VarTSentCount = priv_dev->stat.var_t_sent_count;
                obuf.VarPReceivedCount = priv_dev->stat.var_p_received_count;
                obuf.VarTReceivedCount = priv_dev->stat.var_t_received_count;
                obuf.MsgReceivedCount = priv_dev->stat.msg_received_count;
                obuf.MsgSentCount = priv_dev->stat.msg_sent_count;
                obuf.APUListSentCount = priv_dev->stat.apu_list_sent_count;
                obuf.APNListSentCount = priv_dev->stat.apn_list_sent_count;
                obuf.BaActivityCount = priv_dev->stat.ba_activity_count;
                obuf.BaStopOnTimeoutCount = priv_dev->stat.ba_stop_on_timeout_count;
                obuf.BaStopOnAnomaliesCount = priv_dev->stat.ba_stop_on_anomalies_count;
                obuf.BaStopByUserCount = priv_dev->stat.ba_stop_by_user_count;
                obuf.BaIdleCount = priv_dev->stat.ba_idle_count;
                obuf.BaSuspendCount = priv_dev->stat.ba_suspend_count;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
                if (!access_ok(args.argIO_Out, sizeof(EDA_GET_STATS_OUTPUT)))
#else
                if (!access_ok(VERIFY_WRITE, args.argIO_Out, sizeof(EDA_GET_STATS_OUTPUT)))
#endif
                    return -EFAULT;

                if (copy_to_user((u8*) args.argIO_Out, (u8*) &obuf, sizeof(EDA_GET_STATS_OUTPUT)))
                    return -EFAULT;

                io_status = sizeof(EDA_GET_STATS_OUTPUT);
                status = STATUS_SUCCESS;

            } break;

        // IOCTL FCA (FipCode Access) - related to FullFip IOSpace
        // --- System Management    
        case IOCTL_FCA_START: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_start(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_TST_WFIFO: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_tst_wfifo(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_TST_RFIFO: // O
            spin_lock(&priv_dev->lock);
            status = fca_tst_rfifo(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_TST_IRQ: // O
            spin_lock(&priv_dev->lock);
            status = fca_tst_irq(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_TST_REG: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_tst_reg(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_TST_TIMER: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_tst_timer(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_VALID: // O
            spin_lock(&priv_dev->lock);
            status = fca_valid(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_GET_MEDIUM: // O
            spin_lock(&priv_dev->lock);
            status = fca_get_medium(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_VALID_MEDIUM: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_valid_medium(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_NOISE_MANAGEMENT: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_noise_management(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        // --- Memory Access
        case IOCTL_FCA_W_PHY: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_w_phy(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_R_PHY: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_r_phy(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_W_WORD: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_w_word(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_R_WORD: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_r_word(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_W_DESC: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_w_desc(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        // --- Variable Access
        case IOCTL_FCA_W_VAR: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_w_var(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_R_VAR: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_r_var(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_R_VAR_SVF: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_r_var_svf(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_R_VAR_TIME: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_r_var_time(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_R_LPRESENT: // O
            spin_lock(&priv_dev->lock);
            status = fca_r_lpresent(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_SEND_APER: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_send_aper(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_PURGE_APER: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_purge_aper(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        // --- Event Access
        case IOCTL_FCA_R_EVT: // O
            spin_lock(&priv_dev->lock);
            status = fca_r_evt(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_ACQ_EOC: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_acq_eoc(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        // --- Message Access
        case IOCTL_FCA_W_MSG: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_w_msg(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_R_MSG: // O
            spin_lock(&priv_dev->lock);
            status = fca_r_msg(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_PURGE_MSG: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_purge_msg(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        // --- Ba Access
        case IOCTL_FCA_START_BA: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_start_ba(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_STOP_BA: // O
            spin_lock(&priv_dev->lock);
            status = fca_stop_ba(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_CONTINUE_BA: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_continue_ba(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_CHANGE_MAC: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_change_mac(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_SYNC_BA: // O
            spin_lock(&priv_dev->lock);
            status = fca_sync_ba(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        case IOCTL_FCA_CHANGE_PARAM: // I + O
            spin_lock(&priv_dev->lock);
            status = fca_change_param(priv_dev, ibuf_sz, obuf_sz, ibuf_ptr, obuf_ptr, &io_status);
            spin_unlock(&priv_dev->lock);
            break;
        default:
            printk(KERN_ERR DRV_HDR "Unsupported IOCTL command (0x%x)\n", cmd);
            status = -EINVAL;
            break;
    }

    if (ibuf_ptr != NULL)
        kfree(ibuf_ptr);

    if (args_ptr != NULL) {
        if (put_user(io_status, (u32*) (&args_ptr->io_status)))
            return -EFAULT;
    }

    if (status != STATUS_SUCCESS)
        status = 0x80000000 + status;

    return status;
}

struct file_operations pcifip_fops = {
    .owner  =   THIS_MODULE,
    .read   =   pcifip_read,
    .write  =   pcifip_write,
    .open   =   pcifip_open,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35))
    .ioctl  = pcifip_ioctl,
#else
    .unlocked_ioctl = pcifip_ioctl,
#endif
    .release =  pcifip_release
};

module_init(pci_init_module);
module_exit(pci_exit_module);
