/*
 * Copyright (C) 2014-2019 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/jiffies.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/uaccess.h>
#include "fa-pci.h"
#include "fa-drv.h"
#include "fa-fip.h"

#include "../include/fa-board.h"

/***************       FCA functions (FipCode Access)       ****************/
// --- System Management
u8 fca_start(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 address = *((u16*) ibuf_ptr);
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz != sizeof(u16)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &address + 1));
    write_register_u8(base_addr, reg->key_l, *(u8*) &address);

    // Send CMDE_START command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_START);
 
    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS !=
        wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY  , TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }
    
    // Read VAR_STATE
    buffer = read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_tst_wfifo(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 size = 0;
    u8  buffer = 0x00;
    u16 loopNumber;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz != ( sizeof(u8) * 8)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_TST_WFIFO command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_TST_WFIFO);

    // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);
                                                         
    if (size != sizeof(u8) * 8) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Writing loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, writing to FILE
        write_register_u8(base_addr, reg->file, *((u8*) ibuf_ptr + loopNumber)); 
    }

    // Wait FR = 0 in USTATE
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_FR, TIME_OUT_DELAY, 0)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }
    
    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // Read USTATE
    buffer = read_register_u8(base_addr, reg->ustate);

    // USTATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;  
}

u8 fca_tst_rfifo(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 size = 0;
    u8 buffer = 0x00;
    u8 outBuffer[ sizeof(u8) * 2] = {0};
    u16 loopNumber;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != ( sizeof(u8) * 2)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_TST_RFIFO command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_TST_RFIFO);

    // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);
                                                         
    if (size != sizeof(u8) * 2) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        outBuffer[loopNumber] = read_register_u8(base_addr, reg->file);
    }

    // Wait FR = 0 in USTATE
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_FR, TIME_OUT_DELAY, 0)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }
    
    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // profil test return (2 bytes) - To operate in user mode
    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer[0], sizeof(u16)))
        return -EFAULT;

    *io_status = sizeof(u16);
    return STATUS_SUCCESS;    
}

u8 fca_tst_irq(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != sizeof(u8)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_TST_IRQ command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_TST_IRQ);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read USTATE
    buffer = read_register_u8(base_addr, reg->ustate);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // USTATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;            
}

u8 fca_tst_reg(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 size = 0;
    u8 buffer = 0x00;
    u8 outBuffer[ sizeof(u8) * 2] = {0};
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (ibuf_sz != ( sizeof(u8) * 2)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *(ibuf_ptr + 1));
    write_register_u8(base_addr, reg->key_l, *ibuf_ptr);

    // Send CMDE_TST_REG command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_TST_REG);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);

    if (size != sizeof(u8) * 2) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Read VAR_STATE & UFLAGS
    outBuffer[1] = read_register_u8(base_addr, reg->var_state);
    outBuffer[0] = read_register_u8(base_addr, reg->uflags);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE & UFLAGS registries in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8) * 2))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8) * 2))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer[0], sizeof(u8) * 2))
        return -EFAULT;

    *io_status = sizeof(u8) * 2;
    return STATUS_SUCCESS;
}

u8 fca_tst_timer(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 size = 0;
    u8 buffer = 0x00;
    u16 loopNumber;
    u8 outBuffer[ sizeof(u8) * 2] = {0};
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != ( sizeof(u8) * 2)) ||
        (obuf_sz != ( sizeof(u8) * 2))) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->time_h, *(ibuf_ptr + 1));
    write_register_u8(base_addr, reg->time_l, *ibuf_ptr);

    // Send CMDE_TST_TIME command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_TST_TIME);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);

    if (size != sizeof(u8) * 2) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        outBuffer[loopNumber] = read_register_u8(base_addr, reg->file);
    }

    // Wait USTATE with BUSY = SV = EOC = 1
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY | MSK_USTATE_SV | MSK_USTATE_EOC , TIME_OUT_DELAY, 1))
        return STATUS_TIMEOUT;

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE & UFLAGS registries in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8) * 2))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8) * 2))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer[0], sizeof(u8) * 2))
        return -EFAULT;

    *io_status = sizeof(u8) * 2;
    return STATUS_SUCCESS;
}

u8 fca_valid(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_VALID_CIRCUIT_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != sizeof(FIP_VALID_CIRCUIT_OUTPUT)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_VALID command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_VALID);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = read_register_u8(base_addr, reg->var_state);
    outBuffer.VarState = (u8) buffer;

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.Uflags = 0x00; // UFLAGS not significant

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_VALID_CIRCUIT_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_VALID_CIRCUIT_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_VALID_CIRCUIT_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_VALID_CIRCUIT_OUTPUT);
        return STATUS_SUCCESS;
    }
    
    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // Wait UFLAGS with STATUS = 1 (firmware opérationnel)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->uflags, MSK_UFLAGS_STATUS, TIME_OUT_DELAY, 1))
        return STATUS_TIMEOUT;

    // Read UFLAGS
    outBuffer.Uflags = read_register_u8(base_addr, reg->uflags);

    // VAR_STATE + UFLAGS return - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_VALID_CIRCUIT_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_VALID_CIRCUIT_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_VALID_CIRCUIT_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_VALID_CIRCUIT_OUTPUT);
    return STATUS_SUCCESS;
}

u8 fca_get_medium(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_GET_MEDIUM_OUTPUT outBuffer = {0}; 
    u8 buffer = 0x00;
    u8 loopNumber;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != sizeof(FIP_GET_MEDIUM_OUTPUT)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_GET_MEDIUM command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_GET_MEDIUM);


    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = read_register_u8(base_addr, reg->var_state);
    outBuffer.VarState = (u8) buffer;

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_GET_MEDIUM_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_GET_MEDIUM_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_GET_MEDIUM_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_GET_MEDIUM_OUTPUT);
        return STATUS_SUCCESS;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < FIP_NETWORK_STATUS_IN_BYTE; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        outBuffer.Data[loopNumber] = read_register_u8(base_addr, reg->file);
    }

    // Read VAR_STATE
    outBuffer.VarState = (u8) read_register_u8(base_addr, reg->var_state);
            
    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_GET_MEDIUM_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_GET_MEDIUM_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_GET_MEDIUM_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_GET_MEDIUM_OUTPUT);
    return STATUS_SUCCESS;
}

u8 fca_valid_medium(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 fielDualCmd;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(u16)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    fielDualCmd = *((u16*) ibuf_ptr);

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &fielDualCmd + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &fielDualCmd));

    // Send CMDE_VALID_MEDIUM command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_VALID_MEDIUM);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);
            
    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_noise_management(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 fielDualCmd;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(u16)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    fielDualCmd = *((u16*) ibuf_ptr);

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &fielDualCmd + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &fielDualCmd));

    // Send CMDE_GES_BRUIT command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_GES_BRUIT);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);
            
    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

// --- Memory Access

u8 fca_w_phy(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_WRITE_PHY_INPUT* pIn;
    u8 buffer = 0x00;
    u8 loopNumber;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(FIP_WRITE_PHY_INPUT)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    pIn = (FIP_WRITE_PHY_INPUT*) ibuf_ptr;

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &pIn->PageNumber + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &pIn->PageNumber));

    // Send CMDE_W_PHY command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_W_PHY);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(u8)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
            return -EFAULT;

        if (put_user(buffer, (u8*) obuf_ptr))
            return -EFAULT;

        *io_status = sizeof(u8);
        return STATUS_SUCCESS;
    }

    // Writing loop
    for (loopNumber = 0; loopNumber < NB_BYTE_PER_SEGMENT; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, writing to FILE
        write_register_u8(base_addr, reg->file, pIn->Data[loopNumber]); 
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_r_phy(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 pageNumber;
    FIP_READ_PHY_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    u8 loopNumber;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(u16)) ||
        (obuf_sz != sizeof(FIP_READ_PHY_OUTPUT))) {
        return STATUS_INVALID_PARAMETER;
    }

    pageNumber = *((u16*) ibuf_ptr);

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &pageNumber + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &pageNumber));

    // Send CMDE_R_PHY command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_R_PHY);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_READ_PHY_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_PHY_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_PHY_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_READ_PHY_OUTPUT);
        return STATUS_SUCCESS;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < NB_BYTE_PER_SEGMENT; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        outBuffer.Data[loopNumber] = read_register_u8(base_addr, reg->file);
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_READ_PHY_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_PHY_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_PHY_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_READ_PHY_OUTPUT);

    return STATUS_SUCCESS;  
}

u8 fca_w_word(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_WRITE_WORD_INPUT* pIn;
    u8 buffer = 0x00;
    u8 loopNumber;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(FIP_WRITE_WORD_INPUT)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    pIn = (FIP_WRITE_WORD_INPUT*) ibuf_ptr;

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->uswitch, *((u8*) &pIn->Address + 2));
    write_register_u8(base_addr, reg->key_h, *((u8*) &pIn->Address + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &pIn->Address));

    // Send CMDE_W_WORD command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_W_WORD);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(u8)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
            return -EFAULT;

        if (put_user(buffer, (u8*) obuf_ptr))
            return -EFAULT;

        *io_status = sizeof(u8);
        return STATUS_SUCCESS;
    }

    // Writing loop
    for (loopNumber = 0; loopNumber < sizeof(u16); loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, writing to FILE
        write_register_u8(base_addr, reg->file, *((u8*) &pIn->Data + loopNumber)); 
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_r_word(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u32 address;
    FIP_READ_WORD_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    u8 loopNumber;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(u32)) ||
        (obuf_sz != sizeof(FIP_READ_WORD_OUTPUT))) {
        return STATUS_INVALID_PARAMETER;
    }

    address = *((u32*) ibuf_ptr);

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->uswitch, *((u8*) &address + 2));
    write_register_u8(base_addr, reg->key_h, *((u8*) &address + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &address));

    // Send CMDE_R_WORD command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_R_WORD);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_READ_WORD_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_WORD_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_WORD_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_READ_WORD_OUTPUT);
        return STATUS_SUCCESS;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < sizeof(u16); loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        *((u8*) &outBuffer.Data + loopNumber)  = read_register_u8(base_addr, reg->file);
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_READ_WORD_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_WORD_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_WORD_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_READ_WORD_OUTPUT);

    return STATUS_SUCCESS;   
}

u8 fca_w_desc(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_WRITE_DESC_INPUT* pIn;
    u8 buffer = 0x00;
    u8 loopNumber;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(FIP_WRITE_DESC_INPUT)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    pIn = (FIP_WRITE_DESC_INPUT*) ibuf_ptr;

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->uswitch, *((u8*) &pIn->Address + 2));
    write_register_u8(base_addr, reg->key_h, *((u8*) &pIn->Address + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &pIn->Address));

    // Send CMDE_W_DESC command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_W_DESC);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(u8)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
            return -EFAULT;

        if (put_user(buffer, (u8*) obuf_ptr))
            return -EFAULT;

        *io_status = sizeof(u8);
        return STATUS_SUCCESS;
    }

    // Writing loop
    for (loopNumber = 0; loopNumber < DESCRIPTOR_SIZE_IN_BYTE; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, writing to FILE
        write_register_u8(base_addr, reg->file, pIn->Data[loopNumber]); 
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

// --- Variable Access

u8 fca_w_var(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_WRITE_VAR_INPUT* pIn;
    u8 buffer = 0x00;
    u16 loopNumber;
    u16 size = 0;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(FIP_WRITE_VAR_INPUT)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    pIn = (FIP_WRITE_VAR_INPUT*) ibuf_ptr;

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &pIn->AccessKey + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &pIn->AccessKey));

    // Send CMDE_W_VAR command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_W_VAR);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(u8)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
            return -EFAULT;

        if (put_user(buffer, (u8*) obuf_ptr))
            return -EFAULT;

        *io_status = sizeof(u8);
        return STATUS_SUCCESS;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);
                                                         
    if (size > (NB_BYTE_PER_SEGMENT - 2)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Writing loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, writing to FILE
        write_register_u8(base_addr, reg->file, pIn->Data[loopNumber]); 
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_r_var(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 accessKey;
    FIP_READ_VAR_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    u16 loopNumber;
    u16 size = 0;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(u16)) ||
        (obuf_sz != sizeof(FIP_READ_VAR_OUTPUT))) {
        return STATUS_INVALID_PARAMETER;
    }

    accessKey = *((u16*) ibuf_ptr);

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &accessKey + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &accessKey));

    // Send CMDE_R_VAR command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_R_VAR);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    
    //-- wait_for_US(FULLFIP_WAIT);

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_READ_VAR_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_VAR_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_VAR_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_READ_VAR_OUTPUT);
        return STATUS_SUCCESS;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);

    outBuffer.Size16 = size;
                                                         
    if (size > NB_BYTE_PER_SEGMENT) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        outBuffer.Data[loopNumber]  = read_register_u8(base_addr, reg->file);
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);
    outBuffer.VarState = (u8) buffer;

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_READ_VAR_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_VAR_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_VAR_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_READ_VAR_OUTPUT);
    return STATUS_SUCCESS;
}

u8 fca_r_var_svf(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 accessKey;
    FIP_READ_VAR_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    u16 loopNumber;
    u16 size = 0;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(u16)) ||
        (obuf_sz != sizeof(FIP_READ_VAR_OUTPUT))) {
        return STATUS_INVALID_PARAMETER;
    }

    accessKey = *((u16*) ibuf_ptr);

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &accessKey + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &accessKey));

    // Send CMDE_R_VAR_SVF command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_R_VAR_SVF);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_READ_VAR_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_VAR_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_VAR_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_READ_VAR_OUTPUT);
        return STATUS_SUCCESS;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);

    outBuffer.Size16 = size;
                                                         
    if (size > NB_BYTE_PER_SEGMENT) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        outBuffer.Data[loopNumber]  = read_register_u8(base_addr, reg->file);
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);
    outBuffer.VarState = (u8) buffer;

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_READ_VAR_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_VAR_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_VAR_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_READ_VAR_OUTPUT);
    return STATUS_SUCCESS;
}

u8 fca_r_var_time(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 accessKey;
    FIP_READ_VAR_TIME_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    u16 loopNumber;
    u16 size = 0;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(u16)) ||
        (obuf_sz != sizeof(FIP_READ_VAR_TIME_OUTPUT))) {
        return STATUS_INVALID_PARAMETER;
    }

    accessKey = *((u16*) ibuf_ptr);

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &accessKey + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &accessKey));

    // Send CMDE_R_VAR_TIME command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_R_VAR_TIME);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_READ_VAR_TIME_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_VAR_TIME_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_VAR_TIME_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_READ_VAR_TIME_OUTPUT);
        return STATUS_SUCCESS;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);

    outBuffer.Size16 = size;
                                                         
    if (size > (NB_BYTE_PER_SEGMENT + 2)) { // ConsumptionTime adding
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        if (loopNumber == 0 || loopNumber == 1) { // ConsumptionTime recovery here
            *((u8*) &outBuffer.ConsumptionTime + loopNumber) = read_register_u8(base_addr, reg->file);
        } else {
            outBuffer.Data[loopNumber - 2]  = read_register_u8(base_addr, reg->file);
        }
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);
    outBuffer.VarState = (u8) buffer;

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_READ_VAR_TIME_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_VAR_TIME_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_VAR_TIME_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_READ_VAR_TIME_OUTPUT);
    return STATUS_SUCCESS;
}

u8 fca_r_lpresent(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_READ_LSTPRESENT_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    u16 loopNumber;
    u16 size = 0;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != sizeof(FIP_READ_LSTPRESENT_OUTPUT)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_R_LPRESENT command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_R_LPRESENT);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_READ_LSTPRESENT_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_LSTPRESENT_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_LSTPRESENT_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_READ_LSTPRESENT_OUTPUT);
        return STATUS_SUCCESS;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);
                                              
    if (size > ( sizeof(FIP_READ_LSTPRESENT_OUTPUT) - 1)) { // Without VarState
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        *((u8*) &outBuffer.Pdu + loopNumber) = read_register_u8(base_addr, reg->file);
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);
    outBuffer.VarState = (u8) buffer;

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_READ_LSTPRESENT_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_LSTPRESENT_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_LSTPRESENT_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_READ_LSTPRESENT_OUTPUT);
    return STATUS_SUCCESS;
}

u8 fca_send_aper(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_SEND_APER_INPUT* pIn;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(FIP_SEND_APER_INPUT)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    pIn = (FIP_SEND_APER_INPUT*) ibuf_ptr;

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &pIn->AccessKey + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &pIn->AccessKey));

    // Send CMDE_SEND_APN or CMDE_SEND_APU command to UCOM
    write_register_u8(base_addr, reg->ucom, pIn->TypeCmd);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_purge_aper(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u8 fifoType;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(u8)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    fifoType = *((u8*) ibuf_ptr);

    if (fifoType != FIFO_URGENT)
        fifoType = FIFO_NORMAL;

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->uswitch, fifoType);

    // Send CMDE_PURGE_APER command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_PURGE_APER);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

// --- Event Access

u8 fca_r_evt(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_READ_EVT_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    u16 loopNumber;
    u16 eventCode;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != sizeof(FIP_READ_EVT_OUTPUT)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_R_EVT command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_R_EVT);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_READ_EVT_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_EVT_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_EVT_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_READ_EVT_OUTPUT);
        return STATUS_SUCCESS;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < ( sizeof(u8) * 4); loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        *((u8*) &outBuffer.EventCodeMsb + loopNumber) = read_register_u8(base_addr, reg->file);
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);
    outBuffer.VarState = (u8) buffer;

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // Treat statistics here
    eventCode = (((u16) outBuffer.EventCodeMsb) << 8) + outBuffer.EventCodeLsb;
    switch (eventCode) {
    case EVT_SEND_VAR_P:    dev->stat.var_p_sent_count++; break;
    case EVT_SEND_VAR_T:    dev->stat.var_t_sent_count++; break;
    case EVT_RECEIVE_VAR_P: dev->stat.var_p_received_count++; break;
    case EVT_RECEIVE_VAR_T: dev->stat.var_t_received_count++; break;
    case EVT_RECEIVE_MSG:   dev->stat.msg_received_count++; break;
    case EVT_SEND_MSG:      dev->stat.msg_sent_count++; break;
    case EVT_SEND_APU:      dev->stat.apu_list_sent_count++; break;
    case EVT_SEND_APN:      dev->stat.apn_list_sent_count++; break;
    case EVT_BA_ACTIVITY:   dev->stat.ba_activity_count++; break;
    case EVT_BA_STOP1:      dev->stat.ba_stop_on_timeout_count++; break;
    case EVT_BA_STOP2:      dev->stat.ba_stop_on_anomalies_count++; break;
    case EVT_BA_STOP3:      dev->stat.ba_stop_by_user_count++; break;
    case EVT_BA_IDLE:       dev->stat.ba_idle_count++; break;
    default:
        if ((eventCode & 0x0500) == 0x0500)
            dev->stat.ba_suspend_count++;
    break;
    }

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_READ_EVT_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_EVT_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_EVT_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_READ_EVT_OUTPUT);
    return STATUS_SUCCESS;
}

u8 fca_acq_eoc(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 acqEoc;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(u16)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    acqEoc = *((u16*) ibuf_ptr);

    if (acqEoc != ACK_POS_EOC)
        acqEoc = ACK_NEG_EOC;

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &acqEoc + 1));
    write_register_u8(base_addr, reg->key_l, *(u8*) &acqEoc);

    // Send CMDE_ACQ_EOC command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_ACQ_EOC);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

// --- Message Access

u8 fca_w_msg(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_WRITE_MSG_INPUT* pIn;
    u8 buffer = 0x00;
    u16 loopNumber;
    u16 size = 0;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != sizeof(FIP_WRITE_MSG_INPUT)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    pIn = (FIP_WRITE_MSG_INPUT*) ibuf_ptr;

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &pIn->ChannelNumber + 1));
    write_register_u8(base_addr, reg->key_l, *((u8*) &pIn->ChannelNumber));

    // Send CMDE_W_MSG command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_W_MSG);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(u8)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
            return -EFAULT;

        if (put_user(buffer, (u8*) obuf_ptr))
            return -EFAULT;

        *io_status = sizeof(u8);
        return STATUS_SUCCESS;
    }

    *((u8*) &size + 1)  = pIn->MsgLengthMsb;
    *((u8*) &size)      = pIn->MsgLengthLsb;
                 
    // Added 4 bytes for the message characteristics:
    // MsgMode + ReservedByte + MsgLengthMsb + MsgLengthLsb
    size += 4;

    if (size > MAX_MSG_LENGTH) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Writing loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, writing to FILE
        write_register_u8(base_addr, reg->file, *((u8*) &pIn->MsgMode + loopNumber)); 
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_r_msg(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    FIP_READ_MSG_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    u16 loopNumber;
    u16 size = 0;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != sizeof(FIP_READ_MSG_OUTPUT)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_R_MSG command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_R_MSG);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_READ_MSG_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_MSG_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_MSG_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_READ_MSG_OUTPUT);
        return STATUS_SUCCESS;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);
                                                         
    if (size > MAX_MSG_LENGTH) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        *((u8*) &outBuffer.MsgReceptionTime + loopNumber)  = read_register_u8(base_addr, reg->file);
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);
    outBuffer.VarState = (u8) buffer;

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_READ_MSG_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_READ_MSG_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_READ_MSG_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_READ_MSG_OUTPUT);
    return STATUS_SUCCESS;
}

u8 fca_purge_msg(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 channelNumber;
    FIP_PURGE_MSG_OUTPUT outBuffer = {0};
    u8 buffer = 0x00;
    u16 loopNumber;
    u16 size = 0;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz != sizeof(u16)) ||
        (obuf_sz != sizeof(FIP_PURGE_MSG_OUTPUT))) {
        return STATUS_INVALID_PARAMETER;
    }

    channelNumber = *((u16*) ibuf_ptr);

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    write_register_u8(base_addr, reg->key_h, *((u8*) &channelNumber + 1));
    write_register_u8(base_addr, reg->key_l, *(u8*) &channelNumber);

    // Send CMDE_PURGE_MSG command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_PURGE_MSG);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
        outBuffer.VarState = (u8) buffer;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(FIP_PURGE_MSG_OUTPUT)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_PURGE_MSG_OUTPUT)))
#endif
            return -EFAULT;

        if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_PURGE_MSG_OUTPUT)))
            return -EFAULT;

        *io_status = sizeof(FIP_PURGE_MSG_OUTPUT);
        return STATUS_SUCCESS;
    }

    *((u8*) &size + 1)  = read_register_u8(base_addr, reg->size16_h);
    *((u8*) &size)      = read_register_u8(base_addr, reg->size16_l);

    outBuffer.BlockNumber = size / 2;
                                                         
    if (outBuffer.BlockNumber > NB_MAX_FIFO_EMISSION_BLOCK) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;
        return STATUS_INVALID_DEVICE_STATE;
    }

    // Reading loop
    for (loopNumber = 0; loopNumber < size; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, reading from FILE
        *((u8*) &outBuffer.BlockFreed + loopNumber)  = read_register_u8(base_addr, reg->file);
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);
    outBuffer.VarState = (u8) buffer;

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(FIP_PURGE_MSG_OUTPUT)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(FIP_PURGE_MSG_OUTPUT)))
#endif
        return -EFAULT;

    if (copy_to_user((u8*) obuf_ptr, (u8*) &outBuffer, sizeof(FIP_PURGE_MSG_OUTPUT)))
        return -EFAULT;

    *io_status = sizeof(FIP_PURGE_MSG_OUTPUT);
    return STATUS_SUCCESS;
}

// --- Ba Access

u8 fca_start_ba(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 baNextMacrocycle;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz != sizeof(u16)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    baNextMacrocycle = *((u16*) ibuf_ptr);

    write_register_u8(base_addr, reg->key_h, *((u8*) &baNextMacrocycle + 1));
    write_register_u8(base_addr, reg->key_l, *(u8*) &baNextMacrocycle);

    // Send CMDE_START_BA command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_START_BA);
 
    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS !=
        wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY  , TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }
    
    // Read VAR_STATE
    buffer = read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_stop_ba(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != sizeof(u8)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_STOP_BA command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_STOP_BA);
 
    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS !=
        wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY  , TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }
    
    // Read VAR_STATE
    buffer = read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_continue_ba(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 baAddrContinue;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz != sizeof(u16)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    baAddrContinue = *((u16*) ibuf_ptr);

    write_register_u8(base_addr, reg->key_h, *((u8*) &baAddrContinue + 1));
    write_register_u8(base_addr, reg->key_l, *(u8*) &baAddrContinue);

    // Send CMDE_CONTINUE_BA command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_CONTINUE_BA);
 
    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS !=
        wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY  , TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }
    
    // Read VAR_STATE
    buffer = read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_change_mac(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 baNextMacrocycle;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz != sizeof(u16)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    baNextMacrocycle = *((u16*) ibuf_ptr);

    write_register_u8(base_addr, reg->key_h, *((u8*) &baNextMacrocycle + 1));
    write_register_u8(base_addr, reg->key_l, *(u8*) &baNextMacrocycle);

    // Send CMDE_CHANGE_MAC command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_CHANGE_MAC);
 
    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS !=
        wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY  , TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }
    
    // Read VAR_STATE
    buffer = read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_sync_ba(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != sizeof(u8)) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_SYNC_BA command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_SYNC_BA);
 
    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS !=
        wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY  , TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }
    
    // Read VAR_STATE
    buffer = read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 fca_change_param(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u8 buffer = 0x00;
    u8 loopNumber;
    *io_status = 0;

    base_addr = dev->IoSpacePciFullFip2.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz  != BA_PARAMETERS_SIZE_IN_BYTE) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    // USTATE Test - Transaction in progress ?
    buffer = read_register_u8(base_addr, reg->ustate);

    if ((buffer & MSK_USTATE_BUSY) == MSK_USTATE_BUSY)
        return STATUS_DEVICE_BUSY;

    // Send CMDE_CHANGE_PARAM command to UCOM
    write_register_u8(base_addr, reg->ucom, CMDE_CHANGE_PARAM);

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    if ((buffer & MSK_VAR_STATE_CNF) == MSK_VAR_STATE_CNF) {

        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

        // Wait USTATE with BUSY = 0 (no transaction in progress)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
            return STATUS_TIMEOUT;

        // If VAR_STATE error, STATUS_SUCCESS is returned
        // VarState error must be exploited in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        if (!access_ok(obuf_ptr, sizeof(u8)))
#else
        if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
            return -EFAULT;

        if (put_user(buffer, (u8*) obuf_ptr))
            return -EFAULT;

        *io_status = sizeof(u8);
        return STATUS_SUCCESS;
    }

    // Writing loop
    for (loopNumber = 0; loopNumber < BA_PARAMETERS_SIZE_IN_BYTE; loopNumber++) {
        // Wait USTATE with BUSY = SV = FR = 1 (transaction in progress + status valid + FIFO ready)
        if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY | MSK_USTATE_FR, TIME_OUT_DELAY, 1)) {
            write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
            return STATUS_TIMEOUT;
        }

        // If USTATE OK, writing to FILE
        write_register_u8(base_addr, reg->file, *((u8*) ibuf_ptr + loopNumber)); 
    }

    // Wait USTATE with BUSY = SV = 1 (transaction in progress + status valid)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_SV | MSK_USTATE_BUSY, TIME_OUT_DELAY, 1)) {
        write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);
        return STATUS_TIMEOUT;
    }

    // Read VAR_STATE
    buffer = (u8) read_register_u8(base_addr, reg->var_state);

    // Send CMDE_CLOSE to UCOM - Closure of the transaction
    write_register_u8(base_addr, reg->ucom, CMDE_CLOSE);

    // Wait USTATE with BUSY = 0 (no transaction in progress)
    if (STATUS_SUCCESS != wait_for(base_addr, reg->ustate, MSK_USTATE_BUSY, TIME_OUT_DELAY, 0))
        return STATUS_TIMEOUT;

    // VAR_STATE registry in output - To operate in user mode
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

/*************** EDA_FIELDUAL functions (Exo Device Access) ****************/
/***************        FielDual in "Compatible mode"       ****************/
u8 eda_fieldual_valid_medium(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u16 FielDualCmd = *((u16*) ibuf_ptr); // Input Parameter
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFieldual.pSpaceBase;
    reg = &dev->reg;

    if ((ibuf_sz != sizeof(u16)) ||
        (obuf_sz != sizeof(u8))) {
        return STATUS_INVALID_PARAMETER;
    }

    // Medium State
    buffer = read_register_u8(base_addr, reg->ch_status);

    if ((FielDualCmd & 0x0005) == 0x0005) {             // Channel 1 & 2 enabled
        write_register_u8(base_addr, reg->enable_ch_1_2, 0xFF);  // no significative value

    } else if ((FielDualCmd & 0x0001) == 0x0001) {      // Channel 1 enabled
        write_register_u8(base_addr, reg->enable_ch_1, 0xFF);        // no significative value

    } else if ((FielDualCmd & 0x0004) == 0x0004) {      // Channel 2 enabled
        write_register_u8(base_addr, reg->enable_ch_2, 0xFF);
    }

    if ((FielDualCmd & 0x0010) == 0x0010) {             // Clear TX Errors
        read_register_u8(base_addr, reg->clr_err);
    }

    if ((FielDualCmd & 0x0020) == 0x0020) {             // Channel 1 reset
        read_register_u8(base_addr, reg->reset_ch_1);
    }

    if ((FielDualCmd & 0x0040) == 0x0040) {             // Channel 2 reset
        read_register_u8(base_addr, reg->reset_ch_2);
    }

    // Medium State
    buffer = read_register_u8(base_addr, reg->ch_status);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;

    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}

u8 eda_fieldual_get_medium(struct pci_priv_device* dev, u32 ibuf_sz, u32 obuf_sz, u8* ibuf_ptr, u8* obuf_ptr, u32* io_status)
{
    void* base_addr = 0;
    struct _reg *reg = NULL;
    u8 buffer = 0x00;
    *io_status = 0;

    base_addr = dev->IoSpacePciFieldual.pSpaceBase;
    reg = &dev->reg;

    if (obuf_sz != sizeof(u8)) {
        return STATUS_INVALID_PARAMETER;
    }

    // Medium State
    buffer = read_register_u8(base_addr, reg->ch_status);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    if (!access_ok(obuf_ptr, sizeof(u8)))
#else
    if (!access_ok(VERIFY_WRITE, obuf_ptr, sizeof(u8)))
#endif
        return -EFAULT;

    if (put_user(buffer, (u8*) obuf_ptr))
        return -EFAULT;
    
    *io_status = sizeof(u8);
    return STATUS_SUCCESS;
}
