/*
 * Copyright (C) 2014-2019 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/param.h>
#include <linux/jiffies.h>
#include <linux/uaccess.h>
#include "fa-pci.h"
#include "../include/fa-board.h"    

#define DELAY 1

u8 read_register_u8(void* base_address, u16 offset)
{
    u8* ptr;
    ptr = (u8*) (((u8*) base_address) + offset);
    return (u8) ioread8((void*) ptr);
}

u32 read_register_u32(void* base_address, u16 offset)
{
    u8* ptr;
    ptr = (u8*) (((u8*) base_address) + offset);
    return (u32) ioread32((void*) ptr);
}

void write_register_u8(void* base_address, u16 offset, u8 val)
{
    u8* ptr;
    ptr = (u8*) (((u8*) base_address) + offset);
    iowrite8(val, (void*) ptr);
}

void write_register_u32(void* base_address, u16 offset, u32 val)
{
    u8* ptr;
    ptr = (u8*) (((u8*) base_address) + offset);
    iowrite32(val, (void*) ptr);
}

void init_fip_infos(struct pci_priv_device *dev)
{
    u32 reg;
    u8 magic_number = 0;

    /* read card_conf register */
    init_card_conf(dev);

    /* init fullfip register offsets. By default we consider a board 
     * based on plx component with a 32-bit alignment.
     * ==> offset = 4
     */
    dev->reg_offset_factor = 4;

    /* by default: 1 fullfip + 1 fieldual + 0 fullfip ram (freeaccess) */
    dev->nb_fullfip = 1;
    dev->nb_fullfip_ram = 0;
    dev->nb_fieldual = 1;

    reg = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, 
            C_REG_PLX_RESOURCES);

    magic_number = (u8) ((reg >> 24) & 0x000000ff);

    if (magic_number == C_REG_MAGIC_NB) {
        /* ressource register is significant */
        if (((reg & C_MK_REG_PCIE_TYPE) >> 14) == 1) {
            /* igloo2 communication => 64-bits alignment */
            dev->reg_offset_factor = 8;
        }
        dev->nb_fullfip = (int) (reg & C_MK_REG_FULLFIP_NB);
        dev->nb_fullfip_ram = (int) ((reg & C_MK_REG_FULLFIP_RAM_NB) >> 8);
        dev->nb_fieldual = (int) ((reg & C_MK_REG_FIELDUAL_NB) >> 4);
    }

    /* registration of registers addresses */
    /* fullfip */
    // r
    dev->reg.ustate = dev->reg_offset_factor * USTATE;
    dev->reg.uflags = dev->reg_offset_factor * UFLAGS;
    dev->reg.var_state = dev->reg_offset_factor * VAR_STATE;
    dev->reg.size16_l = dev->reg_offset_factor * SIZE16_L;
    dev->reg.size16_h = dev->reg_offset_factor * SIZE16_H;
    // w
    dev->reg.ucom = dev->reg_offset_factor * UCOM;
    dev->reg.uswitch = dev->reg_offset_factor * USWITCH;
    dev->reg.key_h = dev->reg_offset_factor * KEY_H;
    dev->reg.key_l = dev->reg_offset_factor * KEY_L;
    dev->reg.time_h = dev->reg_offset_factor * TIME_H;
    dev->reg.time_l = dev->reg_offset_factor * TIME_L;
    // r/w
    dev->reg.file = dev->reg_offset_factor * FILE_REG;
    /* fieldual */
    // r
    dev->reg.ch_status = dev->reg_offset_factor * CHANNEL_STATUS;
    dev->reg.clr_err = dev->reg_offset_factor * CLEAR_ERROR;
    dev->reg.reset_ch_1 = dev->reg_offset_factor * RESET_CHANNEL1;
    dev->reg.reset_ch_2 = dev->reg_offset_factor * RESET_CHANNEL2;
    // w
    dev->reg.loopback_sel = dev->reg_offset_factor * LOOPBACK_SELECTION;
    dev->reg.enable_ch_1 = dev->reg_offset_factor * ENABLE_CHANNEL1;
    dev->reg.enable_ch_2 = dev->reg_offset_factor * ENABLE_CHANNEL2;
    dev->reg.enable_ch_1_2 = dev->reg_offset_factor * ENABLE_CHANNEL1_AND_2;
}

void init_card_conf(struct pci_priv_device *dev)
{
    dev->CardConf = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CS3BASE);
}

void init_interrupt(struct pci_priv_device *dev)
{
    u32 data;
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR);
    /* enable LINTi1 (IRQn) and LINTi2 (EOC) */
    data |= C_MK_REG_BIT_IT_LINT1_ENABLE;
    data |= C_MK_REG_BIT_IT_LINT2_ENABLE;
    /* LINTi1 (IRQn) - polarity 0: active low */
    data &= ~C_MK_REG_BIT_IT_LINT1_POLARITY_1;
    /* LINTi2 (EOC) - polarity 1: active high */
    data |= C_MK_REG_BIT_IT_LINT2_POLARITY_1;
    /* LINTi1 (IRQn) - level triggerable interrupt */
    data &= ~C_MK_REG_BIT_IT_LINT1_EDGE_ENABLE;
    /* LINTi2 (EOC) - edge triggerable interrupt */ 
    data |= C_MK_REG_BIT_IT_LINT2_EDGE_ENABLE;
    /* local edge triggerable interrupt clear*/
    data |= C_MK_REG_BIT_IT_LINT1_EDGE_CLEAR;
    data |= C_MK_REG_BIT_IT_LINT2_EDGE_CLEAR;
    /* PCI interrupt enable */
    data |= C_MK_REG_BIT_IT_ENABLE;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR, data);
}

void reset_eoc_interrupt(struct pci_priv_device *dev)
{
    u32 data;
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR);
    /* local edge triggerable interrupt clear*/
    data |= C_MK_REG_BIT_IT_LINT2_EDGE_CLEAR;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_INTCSR, data);
}

void plx_eeprom_enable(struct pci_priv_device *dev)
{
    u32 data;
    wait_ns(DELAY);
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data |= C_MK_REG_BIT_EEPROM_CS_HIGH;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, data);
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
}

void plx_eeprom_disable(struct pci_priv_device *dev)
{ 
    u32 data;
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data &= ~C_MK_REG_BIT_EEPROM_CS_HIGH;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, data);    
}

u8 plx_eeprom_read(struct pci_priv_device *dev, u8* buffer, u8 start_address, u8 length)
{   
    // start_address : Address of the 1st 16-bit word to read
    // length : Amount of 16-bit words to read
    u32 data;
    u8 mask, counter, i, j, k;
    u8 jo;

    counter = 0;
    plx_eeprom_enable(dev);

    plx_eeprom_write_data_bit_high(dev); /* Q = 1 */
    plx_eeprom_write_data_bit_high(dev); /* Q = 1 */
    plx_eeprom_write_data_bit_low(dev); /* Q = 0 */
    
    if ((dev->CardConf & C_MK_REG_PLX_TYPE) == 0x00000000) {
        /* plx 9052 */
        /* we pull down the "A7-A6" bits of start_address by security
         * because only the 6 lower bits should be send */
        start_address &= 0x3F;

        mask = 0x20; /* mask = byte to send (A5 in first) */
        for (k = 0; k < 6; k++) {
            data = start_address & mask;
            (data == mask) ? plx_eeprom_write_data_bit_high(dev) : /* Q = 1 */
                             plx_eeprom_write_data_bit_low(dev);   /* Q = 0 */
            mask /= 2;
        }
    }

    if ((dev->CardConf & C_MK_REG_PLX_TYPE) == 0x00010000) {
        /* plx 9030 */
        mask = 0x80; /* mask = byte to send (A7 in first) */
        for (k = 0; k < 8; k++) {
            data = start_address & mask;
            (data == mask) ? plx_eeprom_write_data_bit_high(dev) : /* Q = 1 */
                             plx_eeprom_write_data_bit_low(dev);   /* Q = 0 */
            mask /= 2;
        }
    }

    /* dummy bit test (should be 0) */
    jo = 0;
    while ((plx_eeprom_read_data_bit_no_clock(dev) != 0) && (jo < 10)) {
        jo++;
    }

    /* reading loop */
    for (i = 0; i < (2 * length); i++) {
        mask = 0x80;
        for (j = 8; j >= 1; j--) {
            data = plx_eeprom_read_data_bit(dev);
            if (data == 0x01) /* bit read is 1 */
                *(buffer + i) |= mask;
            mask /= 2;
        }
        counter++;
    }
    plx_eeprom_disable(dev);

    return counter;
}

void plx_eeprom_do_clock_cycle(struct pci_priv_device *dev)
{
    u32 data;
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data &= ~C_MK_REG_BIT_EEPROM_CLOCK_HIGH;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, data); 
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data |= C_MK_REG_BIT_EEPROM_CLOCK_HIGH;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, data); 
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data &= ~C_MK_REG_BIT_EEPROM_CLOCK_HIGH;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL,data);         
}   

void plx_eeprom_write_data_bit_high(struct pci_priv_device *dev)
{
    u32 data;
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data |= C_MK_REG_BIT_EEPROM_WRITE_HIGH; 
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, data); 
    plx_eeprom_do_clock_cycle(dev);
}

void  plx_eeprom_write_data_bit_low(struct pci_priv_device *dev)
{
    u32 data;
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data &= ~C_MK_REG_BIT_EEPROM_WRITE_HIGH;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, data); 
    plx_eeprom_do_clock_cycle(dev);
}

u8 plx_eeprom_read_data_bit(struct pci_priv_device *dev)
{
    u32 data;
    u8 result;
    plx_eeprom_do_clock_cycle(dev);
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    result = (data & C_MK_REG_BIT_EEPROM_READ_BIT) ? 0x01 : 0x00;
    return result;
}

u8 plx_eeprom_read_data_bit_no_clock(struct pci_priv_device *dev)
{
    u32 data;
    u8 result;
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    result = (data & C_MK_REG_BIT_EEPROM_READ_BIT) ? 0x01 : 0x00;
    return result;
}

void plx_eeprom_reload(struct pci_priv_device*dev)
{
    u32 data;
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data &= ~C_MK_REG_BIT_EEPROM_RELOAD_HIGH;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, data); 
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data |= C_MK_REG_BIT_EEPROM_RELOAD_HIGH;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, data); 
    data = read_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL);
    data &= ~C_MK_REG_BIT_EEPROM_RELOAD_HIGH;
    write_register_u32(dev->IoSpacePciPlx.pSpaceBase, C_REG_PLX_CNTRL, data); 
}

void wait_ns(unsigned long delay)
{
    unsigned long cur_jiffies;
    unsigned long stop;
    cur_jiffies = jiffies;
    stop = cur_jiffies + delay;
    while (stop > cur_jiffies)
        cur_jiffies = jiffies;
}

void wait_us(unsigned long delay)
{
    unsigned long stop;
    bool wait;
    stop = jiffies + usecs_to_jiffies(delay);
    wait = true;
    do {
        if (time_after(jiffies, stop))
            wait = false;
    } while (wait);
}

u8 wait_for(void* addr, u8 reg, u8 status, u32 timeout_us, u8 trigger_edge)
{
    unsigned long stop;
    u8 status_to_check;
    u8 status_read;
    stop = jiffies + usecs_to_jiffies(timeout_us);
    status_to_check = (trigger_edge == 1) ? status : 0x00;
    status_read = read_register_u8(addr, reg);
    while ((status_read & status) != status_to_check) {
        if (time_after(jiffies, stop))
            return STATUS_TIMEOUT;
        status_read = read_register_u8(addr, reg);
    }
    return STATUS_SUCCESS;
}
